#include <asm/segment.h>

#include <rtl.h>
#include <rtl_fifo.h>
#include <rtl_time.h>
#include <pthread.h>

#include <asm/io.h>
#include <linux/ioport.h>
#include <asm/uaccess.h>

#include "rtl_pcl812pg.h"

struct adc_settings {
	short gain[PCL_ADC_CHANNELS];
	short mode;
};

pthread_t dac_threads[PCL_DAC_CHANNELS];
pthread_t do_threads[PCL_DO_BITS];

//#define DEBUG

__inline unsigned int msec(hrtime_t time)
{
	return (unsigned int)((time*1000000)/HRTICKS_PER_SEC);
}

/* Struktur in der die DAC-Aktionen abgelegt werden */
struct dac_event {
	int event;
	hrtime_t delay;
};	

/* Datenstruktur fr einen DAC-Kanal */
struct dac_data {
	int eventcount;
	int current_event;
	int wait;
	hrtime_t starttime;
	int sweep_step;
	int sweep_endvalue;
	struct dac_event event[PCL_DAC_MAXEVENTS];	
};


/* Datenstruktur fr einen DO-Kanal */
struct do_event {
	int event;
	hrtime_t delay;
};	

struct do_data {
	int eventcount;
	int current_event;
	int wait;
	hrtime_t starttime;
	struct do_event event[PCL_DO_MAXEVENTS];	
};

/* Statische Variablen des Moduls */
/* fr die analoge Ausgabe */
static struct dac_data dacdata[PCL_DAC_CHANNELS];
static short dac_portval[PCL_DAC_CHANNELS];
/* fr die digitale Ausgabe */
static struct do_data dodata[PCL_DO_BITS+1];
static short do_portval;
/* fr die analoge Eingabe */
static struct adc_settings adcsettings;

/********************************************************************\
 *	dac_thread(int channel)												*
 *																	*
\********************************************************************/

void * dac_thread(void * arg) {
	register int channel = (unsigned) arg;
	register int thisevent;
	hrtime_t time=0;
	hrtime_t delay=0;

	#ifdef DEBUG
	hrtime_t now=0;
	printk("rtl_pcl812pg: DAC_thread started\n");
	#endif
	
	while (1) {
		#ifdef DEBUG
		printk("\nrtl_pcl812pg: DAC_thread running\n");
		now = gethrtime();
		#endif
		
		// Ereigniszhler um 1 erhhen
		thisevent = dacdata[channel].current_event++;
		
		if (dacdata[channel].eventcount == 0) {
			#ifdef DEBUG
			printk("rtl_pcl812pg: dac_thread(%d) startup & suspend\n",channel);
			#endif
			pthread_suspend_np(pthread_self());
			#ifdef DEBUG
			printk("rtl_pcl812pg: dac_thread(%d) awake\n",channel);
			#endif
		} else {
			#ifdef DEBUG
			printk("rtl_pcl812pg: dac_thread(%d) event %d of %d (now: %u)\n",
					channel, thisevent+1, dacdata[channel].eventcount, msec(now));
			#endif
			
			// 1. Fall: Ein Sweep luft
			if (dacdata[channel].event[0].event == PCL_DAC_SWEEP) {
				dac_portval[channel] += dacdata[channel].sweep_step;
				outb(dac_portval[channel] & 0xff, PCL812PG_DAC_BASE+(2*channel));
				outb(dac_portval[channel] >> 8  , PCL812PG_DAC_BASE+(2*channel)+1);
				#ifdef DEBUG
				printk("rtl_pcl812pg value: sweepstep %d :hi %d, lo %d\n", 
						thisevent, dac_portval[channel]>>8, dac_portval[channel]&0xff);
				#endif
				
				delay = dacdata[channel].event[0].delay;
				if (thisevent != 0) {
					dacdata[channel].starttime += delay;
				}
				time = dacdata[channel].starttime;

			} else {
				// 2. Fall: Direktes setzen eines 12-Bit Wertes:
				if (dacdata[channel].event[thisevent].event < PCL_DAC_RANGE) {
					dac_portval[channel] = dacdata[channel].event[thisevent].event;
					outb(dac_portval[channel] & 0xff, PCL812PG_DAC_BASE+(2*channel));
					outb(dac_portval[channel] >> 8  , PCL812PG_DAC_BASE+(2*channel)+1);
					#ifdef DEBUG
					printk("rtl_pcl812pg value: hi %d, lo %d\n", 
							dac_portval[channel]>>8, dac_portval[channel]&0xff);
					#endif
				}
				
				// 3. Fall: sonstiges
				switch(dacdata[channel].event[thisevent].event) {
					case PCL_DAC_LEVEL_PULSE: // TODO
						outb(dac_portval[channel], PCL812PG_DAC_BASE);
						outb(dac_portval[channel], PCL812PG_DAC_BASE);
						break;
						
					case PCL_DAC_LEVEL_NOP:
						#ifdef DEBUG
						printk("LEVEL_NOP @ %Lu\n", gethrtime());
						#endif
						break;
				}

				if (thisevent != 0) {
					dacdata[channel].starttime += dacdata[channel].event[thisevent-1].delay;
				}
				time = dacdata[channel].starttime;
				delay = dacdata[channel].event[thisevent].delay;
				
				#ifdef DEBUG
				printk("            dt: %d(%u-%u), command: 0x%x, delay: %u\n",
					msec(now)-msec(time), msec(now), msec(time),
						dacdata[channel].event[thisevent].event,
						msec(dacdata[channel].event[thisevent].delay));
				#endif
			}
			
			if (thisevent == dacdata[channel].eventcount-1) {
				// Triggersequenz ist zu Ende
				dacdata[channel].current_event = 0;
				
				if (dacdata[channel].event[0].event != PCL_DAC_SWEEP) {
					if (delay == 0) {
						// Falls das Ende der Triggersequenz erreicht ist, schlafen
						#ifdef DEBUG
						printk("rtl_pcl812pg: DAC_thread(%d) suspend\n",channel);
						#endif
						pthread_suspend_np(pthread_self()); 
						#ifdef DEBUG
						printk("rtl_pcl812pg: DAC_thread(%d) wakeup\n",channel);
						#endif
					} else {
						// Falls die Triggersequenz periodisch ist wieder
						// von vorne anfangen
						dacdata[channel].starttime += delay;
						#ifdef DEBUG
						printk("rtl_pcl812pg: DAC_thread make_periodic @%u, period %u\n", 
								msec(dacdata[channel].starttime), msec(delay));
						#endif
						pthread_make_periodic_np(dac_threads[channel], dacdata[channel].starttime, delay);
					}
				} else {
					// Jetzt zur Sicherheit nochmal den wahren Endwert setzen:
					dac_portval[channel] = dacdata[channel].sweep_endvalue;
					outb(dac_portval[channel] & 0xff, PCL812PG_DAC_BASE+(2*channel));
					outb(dac_portval[channel] >> 8  , PCL812PG_DAC_BASE+(2*channel)+1);
					
					#ifdef DEBUG
					printk("rtl_pcl812pg: end of sweep (DAC_thread suspend)\n");
					#endif
					pthread_suspend_np(pthread_self());
				}
			} else {
				/* die Wartezeit bis num nchsten Aufruf des Tasks einstellen */

				#ifdef DEBUG
				printk("rtl_pcl812pg: DAC_thread make_periodic @%u, period %u\n", 
						msec(time), msec(delay));
				#endif
				pthread_make_periodic_np(dac_threads[channel], time + delay, delay);
			}
		}
	
		// falls der Task angehalten werden soll ...	
		if (dacdata[channel].wait) {
			#ifdef DEBUG
			printk("rtl_pcl812pg: DAC_thread suspended (bw)\n");
			#endif
			dacdata[channel].wait=0;
			pthread_suspend_np(pthread_self());
			#ifdef DEBUG
			printk("rtl_pcl812pg: DAC_thread awake (bw)\n");
			#endif
		}

		/* bis zum nchsten Aufruf des Tasks schlafen */
		#ifdef DEBUG
		printk("rtl_pcl812pg: DAC_thread wait (%u)\n", msec(gethrtime()));
		#endif
		pthread_wait_np();
		
		// falls der Task angehalten werden soll ...	
		if (dacdata[channel].wait) {
			#ifdef DEBUG
			printk("rtl_pcl812pg: DAC_thread suspended (aw)\n");
			#endif
			dacdata[channel].wait=0;
			pthread_suspend_np(pthread_self());
			#ifdef DEBUG
			printk("rtl_pcl812pg: DAC_thread awake (aw)\n");
			#endif
			
			/* bis zum nchsten Aufruf des Tasks schlafen */
			#ifdef DEBUG
			printk("rtl_pcl812pg: DAC_thread wait (%u) (aw)\n", msec(gethrtime()));
			#endif
			pthread_wait_np();
		}
	}
}

/********************************************************************\
 *	dac_add_event(unsigned int channel, int event, hrtime_t delay)		*
 *																	*
\********************************************************************/

void dac_add_event(unsigned int channel, int event, hrtime_t delay)
{ 
	if (dacdata[channel].eventcount < PCL_DAC_MAXEVENTS) {
		int eventcount = dacdata[channel].eventcount++;
		
		#ifdef DEBUG
/*		printk("rtl_pcl812pg_add_event: #%d, event: 0x%x, delay %u\n",
				eventcount+1, event, msec(delay));*/
		#endif
		
		dacdata[channel].event[eventcount].event = event;
		dacdata[channel].event[eventcount].delay = delay;
	} else {
		printk("rtl_pcl812pg: ERROR - DAC_MAXEVENTS reached!\n");
	}
}

/********************************************************************\
 *	dac_set_sweep(unsigned int channel, hrtime_t sweepdelay,				*
 				  int sweepstep, int sweepcount, int sweepstop)		*
 *																	*
\********************************************************************/

void dac_set_sweep(unsigned int channel, hrtime_t sweepdelay,
		int sweepstep, int sweepcount, int sweepstop)
{
		
	#ifdef DEBUG
	printk("rtl_pcl812pg_set_sweep: delay %u, dx %d, #%d\n",
		msec(sweepdelay), sweepstep, sweepcount);
	#endif
	dacdata[channel].event[0].event = PCL_DAC_SWEEP;
	dacdata[channel].event[0].delay = sweepdelay;
	dacdata[channel].sweep_step     = sweepstep;
	dacdata[channel].sweep_endvalue = sweepstop;
	dacdata[channel].eventcount     = sweepcount;
	dacdata[channel].current_event  = 0;
}

/********************************************************************\
 *	do_thread(int bit)												*
 *																	*
\********************************************************************/

void * do_thread(void * arg) {
	register int bit = (unsigned) arg;
	register int thisevent;
	hrtime_t time=0;
	hrtime_t delay=0;

	#ifdef DEBUG
	hrtime_t now=0;
	printk("rtl_pcl812pg: DO_thread started\n");
	#endif
	
	while (1) {
		#ifdef DEBUG
		printk("\nrtl_pcl812pg: DO_thread running\n");
		now = gethrtime();
		#endif
		
		// Ereigniszhler um 1 erhhen
		thisevent = dodata[bit].current_event++;
		
		if (dodata[bit].eventcount == 0) {
			#ifdef DEBUG
			printk("rtl_pcl812pg: do_thread(%d) startup & suspend\n",bit);
			#endif
			pthread_suspend_np(pthread_self());
		} else {

			#ifdef DEBUG
			printk("rtl_pcl812pg: do_thread(%d) event %d of %d \n",
					bit, thisevent+1, dodata[bit].eventcount);
			#endif
			
			switch(dodata[bit].event[thisevent].event) {
				case PCL_DO_LEVEL_HIGH: 
					do_portval |= (1 << bit);
					outb(do_portval & 0xff, PCL812PG_DO_LO);
					outb(do_portval >> 8, PCL812PG_DO_HI);
					break;
					
				case PCL_DO_LEVEL_LOW:
					do_portval &= ~(1 << bit);
					outb(do_portval & 0xff, PCL812PG_DO_LO);
					outb(do_portval >> 8, PCL812PG_DO_HI);
					break;
					
				case PCL_DO_LEVEL_PULSE:
					do_portval ^= (1 << bit);
					outb(do_portval & 0xff, PCL812PG_DO_LO);
					outb(do_portval >> 8, PCL812PG_DO_HI);
					do_portval ^= (1 << bit);
					outb(do_portval & 0xff, PCL812PG_DO_LO);
					outb(do_portval >> 8, PCL812PG_DO_HI);
					break;
						
				case PCL_DO_LEVEL_NOP:
					#ifdef DEBUG
					printk("LEVEL_NOP @ %Lu\n", gethrtime());
					#endif
					break;
					
				default:
					printk("rtl_pcl812pg: ERROR unknown event %d!\n",
							dodata[bit].event[thisevent].event);
			}

			if (thisevent != 0) {
				dodata[bit].starttime += dodata[bit].event[thisevent-1].delay;
			}
			
			time = dodata[bit].starttime;
			delay = dodata[bit].event[thisevent].delay;
			
			#ifdef DEBUG
			printk("            dt: %d(%u-%u), command: 0x%x, delay: %u\n",
					msec(now)-msec(time), msec(now), msec(time),
					dodata[bit].event[thisevent].event,
					msec(dodata[bit].event[thisevent].delay));
			#endif
			
			if (thisevent == dodata[bit].eventcount-1) {
				// Triggersequenz ist zu Ende
				dodata[bit].current_event = 0;
				
				if (delay == 0) {
					// Falls das Ende der Triggersequenz erreicht ist, schlafen
					pthread_suspend_np(pthread_self()); 
				} else {
					// Falls die Triggersequenz periodisch ist wieder
					// von vorne anfangen
					dodata[bit].starttime += delay;
					pthread_make_periodic_np(do_threads[bit], dodata[bit].starttime, delay);
				}
			} else {
				/* die Wartezeit bis num nchsten Aufruf des Tasks einstellen */

				pthread_make_periodic_np(do_threads[bit], time+delay, delay);
			}
		}
	
		// Falls der Thread angehalten werden soll ...	
		if (dodata[bit].wait) {
			#ifdef DEBUG
			printk("rtl_pcl819pg: DOUT_thread suspended (bw)\n");
			#endif
			dodata[bit].wait=0;
			pthread_suspend_np(pthread_self());
			#ifdef DEBUG
			printk("rtl_pcl812pg: DOUT_thread awake (bw)\n");
			#endif
		}
		
		/* bis zum nchsten Aufruf des Tasks schlafen */
		#ifdef DEBUG
		printk("rtl_pcl812pg: DOUT_thread wait\n");
		#endif
		pthread_wait_np();
		
		// Falls der Thread angehalten werden soll ...	
		if (dodata[bit].wait) {
			#ifdef DEBUG
			printk("rtl_pcl812pg: DOUT_thread suspended (aw)\n");
			#endif
			dodata[bit].wait=0;
			pthread_suspend_np(pthread_self());
			#ifdef DEBUG
			printk("rtl_pcl812pg: DOUT_thread awake (aw)\n");
			#endif
			
			/* bis zum nchsten Aufruf des Tasks schlafen */
			#ifdef DEBUG
			printk("rtl_pcl812pg: DOUT_thread wait (aw)\n");
			#endif
			pthread_wait_np();
		}
	}
}

/********************************************************************\
 *	do_add_event(unsigned int bit, int event, hrtime_t delay)		*
 *																	*
\********************************************************************/

void do_add_event(unsigned int bit, int event, hrtime_t delay)
{ 
	if (dodata[bit].eventcount < PCL_DO_MAXEVENTS) {
		int eventcount = dodata[bit].eventcount++;
		
		#ifdef DEBUG
		printk("rtl_pcl812pg_add_event: #%d, event: 0x%x, delay %u\n",
				eventcount+1, event, msec(delay));
		#endif
		
		dodata[bit].event[eventcount].event = event;
		dodata[bit].event[eventcount].delay = delay;
	} else {
		printk("rtl_pcl812pg: ERROR - PCL_DO_MAXEVENTS reached!\n");
	}
}
	
/********************************************************************\
 *	pcl_fifo_handler(unsigned int fifo)								*
 *																	*
\********************************************************************/

int pcl_fifo_handler(unsigned int fifo)
{
	struct rtl_pcl_msg msg;
	int err=0;
	int i=0, dac_stepsize;
	int dac_sweepstart, dac_sweepstop, dac_sweepdelta, dac_sgn;
	hrtime_t dac_sweeptime, dac_dt=0;
	hrtime_t starttime;

	while ((err = rtf_get(fifo, &msg, sizeof(msg))) == sizeof(msg)) {
		#ifdef DEBUG
		printk("rtl_pcl812pg: chan#%d (cmd: %d, delay: %u, endvalue: %d)\n", msg.channel,
				msg.command, msec(msg.delay), msg.endvalue);
		#endif

		switch (msg.command) {
			
			/* Zuerst die Kommandos der analogen Ausgabe */
			
			case PCL_DAC_SEQUENCE_START:
				starttime = gethrtime() + 0.1 * HRTICKS_PER_SEC;
				
				for (i=0; i<PCL_DAC_CHANNELS; i++) {
					if ((msg.channel & (1 << i))!= 0) {
						if (dacdata[i].eventcount > 0) {
							dacdata[i].starttime = starttime;
							dacdata[i].current_event = 0;
							dacdata[i].wait = 0;
							#ifdef DEBUG
							printk("rtl_pcl812pg: %u, starting dac #%d sequence at %u\n",
									msec(gethrtime()),i,msec(starttime));
							#endif
							pthread_make_periodic_np(dac_threads[i], starttime, 10000);
							pthread_wakeup_np(dac_threads[i]);
						} else {
							#ifdef DEBUG
							printk("rtl_pcl812pg: no dac-events for dac #%d\n", i);
							#endif
						}
					}
				}
				break;
				
			case PCL_DAC_SEQUENCE_STOP:
				#ifdef DEBUG
				printk("rtl_pcl812pg: SEQUENCE_STOP at %u\n", msec(gethrtime()));
				#endif
				
				for (i=0; i<PCL_DAC_CHANNELS; i++) {
					if ((msg.channel & (1 << i))!= 0) {
						dacdata[i].wait=1;
						#ifdef DEBUG
						printk("rtl_pcl812pg: sequence of dac #%d stopped\n", i);
						#endif
					}
				}
				
				break;
				
			case PCL_DAC_SEQUENCE_CLEAR:
				dacdata[msg.channel].wait=1;
				#ifdef DEBUG
				printk("rtl_pcl812pg: SEQUENCE_CLEAR\n");
				#endif
				dacdata[msg.channel].eventcount=0;
				break;
				
			case PCL_DAC_SET:
				dacdata[msg.channel].wait=1;
				dac_portval[msg.channel] = msg.endvalue;
				outb(dac_portval[msg.channel] & 0xff, PCL812PG_DAC_BASE+2*msg.channel);
				outb(dac_portval[msg.channel] >> 8  , PCL812PG_DAC_BASE+2*msg.channel+1);
				#ifdef DEBUG
				printk("rtl_pcl812pg set value: hi %d, lo %d\n",
						dac_portval[msg.channel]>>8, dac_portval[msg.channel]&0xff);
				#endif
				break;
				
			case PCL_DAC_SWEEP:
				dacdata[msg.channel].wait=1;
				dac_sweepstart = dac_portval[msg.channel];
				dac_sweepstop = msg.endvalue;
				dac_sweeptime = msg.delay;
				dac_sweepdelta = abs(dac_sweepstop-dac_sweepstart);
				
				if (dac_sweepdelta==0) {
					#ifdef DEBUG
					printk("rtl_pcl812pg sweep: no sweep (start=stop)\n");
					#endif
					break;
				}
				
				for (dac_stepsize=1; dac_stepsize<dac_sweepdelta; dac_stepsize++) {
					dac_dt = (dac_sweeptime*dac_stepsize)/dac_sweepdelta;
					
					#ifdef DEBUG
					printk("rtl_pcl812pg sweep: steps=%d\n", dac_sweepdelta/dac_stepsize);
					printk("rtl_pcl812pg sweep: dt=%u us\n",
							(msec(dac_dt)*1000));
					printk("%u > %u ?\n", msec(dac_dt),
							msec((HRTICKS_PER_SEC/1000000*PCL_DAC_SWEEP_MINTIME)));
					#endif

					if (dac_dt > (HRTICKS_PER_SEC/1000000.0*PCL_DAC_SWEEP_MINTIME)) {
						#ifdef DEBUG
						printk("rtl_pcl812pg sweep: dx=%d\n", dac_stepsize);
						#endif
						break;
					}
				}
				
				#ifdef DEBUG	
				printk("rtl_pcl812pg sweep: %d steps (dx=%d) with dt=%u us in %u ms\n",
						dac_sweepdelta/dac_stepsize, dac_stepsize,
						msec(dac_dt)*1000,
						msec(dac_sweepdelta*dac_dt/dac_stepsize));
				#endif

				dac_sgn = (dac_sweepstop-dac_sweepstart)/dac_sweepdelta;
				dac_set_sweep(msg.channel, dac_dt, dac_stepsize*dac_sgn,
							  dac_sweepdelta/dac_stepsize, dac_sweepstop);
				
				starttime = gethrtime() + 0.1 * HRTICKS_PER_SEC;
				dacdata[msg.channel].starttime = starttime;
				dacdata[msg.channel].wait = 0;
				pthread_make_periodic_np(dac_threads[msg.channel], starttime, 10000);
				pthread_wakeup_np(dac_threads[msg.channel]);
				#ifdef DEBUG
				printk("rtl_pcl812pg: starting sweep for dac #%d\n", msg.channel);
				#endif
				break;
				
			case PCL_DAC_LEVEL_NOP:
				dac_add_event(msg.channel, msg.command, msg.delay);
				break;
				
			case PCL_DAC_LEVEL_PULSE:
				dac_add_event(msg.channel, msg.command, msg.delay);
				break;
			
			/* Dann die Kommandos der digitalen Ausgabe */
				
			case PCL_DO_SEQUENCE_START:
				starttime = gethrtime() + 0.1 * HRTICKS_PER_SEC;
				
				for (i=0; i<PCL_DO_BITS; i++) {
					if ((msg.channel & (1 << i))!= 0) {
						if (dodata[i].eventcount > 0) {
							dodata[i].starttime = starttime;
							dodata[i].wait = 0;
							dodata[i].current_event = 0;
							#ifdef DEBUG
							printk("rtl_pcl812pg: %u, starting DO_bit #%d sequence at %u\n",
									msec(gethrtime()),i,msec(starttime));
							#endif
							pthread_make_periodic_np(do_threads[i], starttime, 10000);
							pthread_wakeup_np(do_threads[i]);
						} else {
							#ifdef DEBUG
							printk("rtl_pcl812pg: no events for DO-bit #%d\n", i);
							#endif
						}
					}
				}
				break;
				
			case PCL_DO_SEQUENCE_STOP:
				#ifdef DEBUG
				printk("rtl_pcl812pg: SEQUENCE_STOP at %u\n", msec(gethrtime()));
				#endif
				
				for (i=0; i<PCL_DO_BITS; i++) {
					if ((msg.channel & (1 << i))!= 0) {
						dodata[i].wait=1;
						#ifdef DEBUG
						printk("rtl_pcl812pg: sequence of do #%d stopped\n", i);
						#endif
					}
				}
				
				break;
				
			case PCL_DO_SEQUENCE_CLEAR:
				dodata[msg.channel].wait=1;
				#ifdef DEBUG
				printk("rtl_pcl812pg: SEQUENCE_CLEAR\n");
				#endif
				dodata[msg.channel].eventcount=0;
				break;
				
			case PCL_DO_SET:
				#ifdef DEBUG
				printk("rtl_pcl812pg set bits: %x\n", msg.channel);
				#endif
				do_portval |= msg.channel;
				outb(do_portval & 0xff, PCL812PG_DO_LO);
				outb(do_portval >> 8, PCL812PG_DO_HI);
				break;
				
			case PCL_DO_CLEAR:
				#ifdef DEBUG
				printk("rtl_pcl812pg set bits: %x\n", msg.channel);
				#endif
				do_portval &= ~msg.channel;
				outb(do_portval & 0xff, PCL812PG_DO_LO);
				outb(do_portval >> 8, PCL812PG_DO_HI);
				
				break;
				
			case PCL_DO_LEVEL_NOP:
				dodata[msg.channel].wait=1;
				do_add_event(msg.channel, msg.command, msg.delay);
				break;
				
			case PCL_DO_LEVEL_HIGH:
				dodata[msg.channel].wait=1;
				do_add_event(msg.channel, msg.command, msg.delay);
				break;
				
			case PCL_DO_LEVEL_LOW:
				dodata[msg.channel].wait=1;
				do_add_event(msg.channel, msg.command, msg.delay);
				break;
				
			case PCL_DO_LEVEL_PULSE:
				dodata[msg.channel].wait=1;
				do_add_event(msg.channel, msg.command, msg.delay);
				break;

			case PCL_ADC_READ:
				outb(msg.channel, PCL812PG_MUX);
				adcsettings.gain[msg.channel]=(1 >> 1);
				outb(adcsettings.gain[msg.channel],PCL812PG_GAIN);
				outb(0,PCL812PG_SOFTTRIG);
				
				while ((inb(PCL812PG_AI_HI) & PCL812PG_AI_HI_DRDY) != 0) {
/*					current->timeout = 50;
					current->state = TASK_INTERRUPTIBLE;
					schedule();
					current->timeout = 0;*/
				}
				
				{
					int *value_ptr=(int *)msg.endvalue;
					short value;
				
					value = (inb(PCL812PG_AI_HI) << 8) + inb(PCL812PG_AI_LO);
					put_user(value,value_ptr);
				}		
				break;
				
			default:
				if ((msg.command >= 0) && (msg.command < PCL_DAC_RANGE)) {
					/* Falls der Wert im gltigen Bereich des DAC liegt wird er als
					 * DAC-Wert interpretiert */
					dacdata[msg.channel].wait=1;
					dac_add_event(msg.channel, msg.command, msg.delay);
				} else {
					return -EINVAL;
				}
		}
	}
	
	if (err != 0) {
		return -EINVAL;
	}

	return 0;
}

/********************************************************************\
 *	init_module(void)												*
 *																	*
\********************************************************************/

int init_module(void)
{
	int i, base, size;
	
	base = PCL812PG_BASE;
	size = PCL812PG_SIZE;

	if (base == 0)
		return -1;
	if (check_region(base, size) < 0)
		return -1;
	
	printk("rtl_pcl812pg: using ioport 0x%04x\n", base);
	// IO-Adresse reservieren
	request_region(base, size, "rtl_pcl812pg");

	// Fr jeden DAC-Kanal einen RT-Task erzeugen
		for (i=0; i<PCL_DAC_CHANNELS; i++) {
		// Output auf 0 setzen
		outb(0x00, PCL812PG_DAC_BASE + 2*i);
		outb(0x00, PCL812PG_DAC_BASE + 2*i+1);
		
		#ifdef DEBUG
		printk("rtl_pcl812pg: INIT DAC task #%d\n", i);
		#endif
		
		/* Anzahl der Trigger-Events auf 0 setzen */
		dacdata[i].eventcount=0;
		
		/* Wartemodus ausschalten */
		dacdata[i].wait=0;
		
		/* den Trigger-Task erzeugen */
		pthread_create(&dac_threads[i], NULL, dac_thread, (void *)i);
	}
		
	// Fr jeden DO-Kanal einen RT-Task erzeugen
	for (i=0; i<PCL_DO_BITS; i++) {
		// Output auf 0 setzen
		
		#ifdef DEBUG
		printk("rtl_pcl812pg: INIT DO task #%d\n", i);
		#endif
		
		/* Anzahl der Trigger-Events auf 0 setzen */
		dodata[i].eventcount=0;
		
		/* Wartemodus ausschalten */
		dodata[i].wait=0;
		
		/* den Trigger-Task erzeugen */
		pthread_create(&do_threads[i], NULL, do_thread, (void *)i);
	}
		
	#ifdef DEBUG
	printk("rtl_pcl812pg: INIT fifo #%d\n", PCL_FIFO);
	#endif
	
	/* Kommando-FIFO erzeugen */
	rtf_create(PCL_FIFO, 100);
		
	/* den FIFO-Handler deklarieren */ 
	rtf_create_handler(PCL_FIFO, &pcl_fifo_handler);
	
	return 0;
}


/********************************************************************\
 *	cleanup_module(void)											*
 *																	*
\********************************************************************/

void cleanup_module(void)
{
	int i, base, size;
	
	base = PCL812PG_BASE;
	size = PCL812PG_SIZE;

	#ifdef DEBUG	
	printk("rtl_pcl812pg: KILL\n");
	#endif
	
	// Die Fifo des Moduls entfernen
	#ifdef DEBUG
	printk("rtl_pcl812pg: DELETE fifo #%d\n", PCL_FIFO);
	#endif
	rtf_destroy(PCL_FIFO);
	
	for (i=0; i<PCL_DAC_CHANNELS; i++) {
		#ifdef DEBUG
		printk("rtl_pcl812pg: DELETE DAC task #%d\n", i);
		#endif
		
		pthread_delete_np(dac_threads[i]);
	}
	
	for (i=0; i<PCL_DO_BITS; i++) {
		#ifdef DEBUG
		printk("rtl_pcl812pg: DELETE DO task #%d\n", i);
		#endif
		
		pthread_delete_np(do_threads[i]);
	}
	
	release_region(base, size);
}
