// Hardwareadresse der Labcard
#define PCL812PG_BASE	0x220
#define PCL812PG_SIZE	0x0f

// Registeradressen der PCL812PG
#define PCL812PG_CTR0		(PCL812PG_BASE + 0x00)
#define PCL812PG_CTR1		(PCL812PG_BASE + 0x01)
#define PCL812PG_CTR2		(PCL812PG_BASE + 0x02)
#define PCL812PG_CTRCTL		(PCL812PG_BASE + 0x03)

#define PCL812PG_DAC_BASE	(PCL812PG_BASE + 0x04)
#define PCL812PG_DA0_LO		(PCL812PG_BASE + 0x04)
#define PCL812PG_AI_LO		(PCL812PG_BASE + 0x04)

#define PCL812PG_DA0_HI		(PCL812PG_BASE + 0x05)
#define PCL812PG_AI_HI		(PCL812PG_BASE + 0x05)
#define PCL812PG_AI_HI_DRDY (1 << 4)

#define PCL812PG_DA1_LO		(PCL812PG_BASE + 0x06)
#define PCL812PG_DI_LO		(PCL812PG_BASE + 0x06)

#define PCL812PG_DA1_HI		(PCL812PG_BASE + 0x07)
#define PCL812PG_DI_HI		(PCL812PG_BASE + 0x07)

#define PCL812PG_GAIN		(PCL812PG_BASE + 0x09)
#define PCL812PG_MUX		(PCL812PG_BASE + 0x0a)
#define PCL812PG_MODE		(PCL812PG_BASE + 0x0b)
#define PCL812PG_SOFTTRIG	(PCL812PG_BASE + 0x0c)
#define PCL812PG_DO_LO		(PCL812PG_BASE + 0x0d)
#define PCL812PG_DO_HI		(PCL812PG_BASE + 0x0e)

// Fifo-Nummer des PCL
#define PCL_FIFO 0

// Aufloesung des DAC
#define PCL_BITS	12
#define PCL_RANGE	(1<<PCL_BITS)
#define PCL_DAC_RANGE PCL_RANGE
#define PCL_ADC_RANGE PCL_RANGE
#define PCL_RANGEMASK (PCL_RANGE-1)
#define PCL_DAC_RANGEMASK PCL_RANGEMASK
#define PCL_ADC_RANGEMASK PCL_RANGEMASK

// Eingestellter Spannungsbereich des DAC
#define PCL_DAC_ANALOGRANGE 10

// Definitin der DAC-Befehle
#define PCL_DAC_LEVEL_NOP	(PCL_RANGE+1)
#define PCL_DAC_LEVEL_HIGH	PCL_RANGEMASK
#define PCL_DAC_LEVEL_LOW	0x000
#define PCL_DAC_LEVEL_PULSE	(PCL_RANGE+4)

#define PCL_DAC_SEQUENCE_START	(PCL_RANGE+100)
#define PCL_DAC_SEQUENCE_STOP	(PCL_RANGE+101)
#define PCL_DAC_SEQUENCE_CLEAR	(PCL_RANGE+102)
#define PCL_DAC_SET				(PCL_RANGE+103)
#define PCL_DAC_SWEEP			(PCL_RANGE+104)
#define PCL_ADC_READ			(PCL_RANGE+200)
#define PCL_DO_SEQUENCE_START	(PCL_RANGE+300)
#define PCL_DO_SEQUENCE_STOP	(PCL_RANGE+301)
#define PCL_DO_SEQUENCE_CLEAR	(PCL_RANGE+302)
#define PCL_DO_SET				(PCL_RANGE+303)
#define PCL_DO_CLEAR			(PCL_RANGE+304)
#define PCL_DO_LEVEL_NOP		(PCL_RANGE+305)
#define PCL_DO_LEVEL_HIGH		(PCL_RANGE+306)
#define PCL_DO_LEVEL_LOW		(PCL_RANGE+307)
#define PCL_DO_LEVEL_PULSE		(PCL_RANGE+308)

// Anzahl der PCL-Register
#define PCL_DAC_CHANNELS 2
#define PCL_ADC_CHANNELS 16
#define PCL_DO_BITS 16
#define PCL_OUTPUT_MAXEVENTS 200
#define PCL_DAC_MAXEVENTS PCL_OUTPUT_MAXEVENTS
#define PCL_DO_MAXEVENTS PCL_OUTPUT_MAXEVENTS

// Minimale Sweepzeit in us
#define PCL_DAC_SWEEP_MINTIME 1000

struct rtl_pcl_msg {
	short channel;	/* DAC-Kanal (fr Ein- und Ausgabe) */
	int command;	/* DAC-Kommado (siehe oben) */
	hrtime_t delay;	/* Wartezeit nach Ausfhrung von <command> (bei DAC_SWEEP: Sweepdauer) */
	int endvalue;	/* bei PCL_DAC_SET: Neuer Wert des DAC */
					/* bei PCL_DAC_SWEEP: Endwert des DAC nach <delay> ticks */
};
